/*----------------------------------------------------------------------
  PuReMD - Purdue ReaxFF Molecular Dynamics Program

  Copyright (2010) Purdue University
  Hasan Metin Aktulga, haktulga@cs.purdue.edu
  Joseph Fogarty, jcfogart@mail.usf.edu
  Sagar Pandit, pandit@usf.edu
  Ananth Y Grama, ayg@cs.purdue.edu

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License as
  published by the Free Software Foundation; either version 2 of
  the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  See the GNU General Public License for more details:
  <http://www.gnu.org/licenses/>.
  ----------------------------------------------------------------------*/

#include "reax_types.h"

#if defined(PURE_REAX)
  #include "tool_box.h"
#elif defined(LAMMPS_REAX)
  #include "reax_tool_box.h"
#endif


/************** taken from comm_tools.c **************/
int SumScan( int n, int me, int root, MPI_Comm comm )
{
    int i, my_order, wsize;
    int *nbuf = NULL;

    if ( me == root )
    {
        MPI_Comm_size( comm, &wsize );
        nbuf = (int *) scalloc( wsize, sizeof(int), "SumScan:nbuf" );

        MPI_Gather( &n, 1, MPI_INT, nbuf, 1, MPI_INT, root, comm );

        for ( i = 0; i < wsize - 1; ++i )
        {
            nbuf[i + 1] += nbuf[i];
        }

        MPI_Scatter( nbuf, 1, MPI_INT, &my_order, 1, MPI_INT, root, comm );

        sfree( nbuf, "SumScan:nbuf" );
    }
    else
    {
        MPI_Gather( &n, 1, MPI_INT, nbuf, 1, MPI_INT, root, comm );
        MPI_Scatter( nbuf, 1, MPI_INT, &my_order, 1, MPI_INT, root, comm );
    }

    return my_order;
}


void SumScanB( int n, int me, int wsize, int root, MPI_Comm comm, int *nbuf )
{
    int i;

    MPI_Gather( &n, 1, MPI_INT, nbuf, 1, MPI_INT, root, comm );

    if ( me == root )
    {
        for ( i = 0; i < wsize - 1; ++i )
        {
            nbuf[i + 1] += nbuf[i];
        }
    }

    MPI_Bcast( nbuf, wsize, MPI_INT, root, comm );
}


/************** taken from box.c **************/
void Transform( rvec x1, simulation_box *box, char flag, rvec x2 )
{
    int i, j;
    real tmp;

#if defined(DEBUG)
    fprintf( stderr, ">x1: (%lf, %lf, %lf)\n",x1[0],x1[1],x1[2] );
#endif

    if ( flag > 0 )
    {
        for (i = 0; i < 3; i++)
        {
            tmp = 0.0;
            for (j = 0; j < 3; j++)
            {
                tmp += box->trans[i][j] * x1[j];
            }
            x2[i] = tmp;
        }
    }
    else
    {
        for (i = 0; i < 3; i++)
        {
            tmp = 0.0;
            for (j = 0; j < 3; j++)
            {
                tmp += box->trans_inv[i][j] * x1[j];
            }
            x2[i] = tmp;
        }
    }

#if defined(DEBUG)
    fprintf( stderr, ">x2: (%lf, %lf, %lf)\n", x2[0], x2[1], x2[2] );
#endif
}


void Transform_to_UnitBox( rvec x1, simulation_box *box, char flag, rvec x2 )
{
    Transform( x1, box, flag, x2 );

    x2[0] /= box->box_norms[0];
    x2[1] /= box->box_norms[1];
    x2[2] /= box->box_norms[2];
}


/* determine whether point p is inside the box */
void Fit_to_Periodic_Box( simulation_box *box, rvec *p )
{
    int i;

    for ( i = 0; i < 3; ++i )
    {
        if ( (*p)[i] < box->min[i] )
        {
            /* handle lower coords */
            while ( (*p)[i] < box->min[i] )
            {
                (*p)[i] += box->box_norms[i];
            }
        }
        else if ( (*p)[i] >= box->max[i] )
        {
            /* handle higher coords */
            while ( (*p)[i] >= box->max[i] )
            {
                (*p)[i] -= box->box_norms[i];
            }
        }
    }
}


/************** from geo_tools.c *****************/
void Make_Point( real x, real y, real z, rvec* p )
{
    (*p)[0] = x;
    (*p)[1] = y;
    (*p)[2] = z;
}


int is_Valid_Serial( storage *workspace, int serial )
{
//    if( workspace->map_serials[ serial ] < 0 )
//    {
//        fprintf( stderr, "CONECT line includes invalid pdb serial number %d.\n",
//                serial );
//        fprintf( stderr, "Please correct the input file.Terminating...\n" );
//        MPI_Abort( MPI_COMM_WORLD, INVALID_INPUT );
//    }

    return SUCCESS;
}


int Check_Input_Range( int val, int lo, int hi, char *message )
{
    if ( val < lo || val > hi )
    {
        fprintf( stderr, "%s\nInput %d - Out of range %d-%d. Terminating...\n",
                 message, val, lo, hi );
        MPI_Abort( MPI_COMM_WORLD, INVALID_INPUT );
    }

    return 1;
}


void Trim_Spaces( char *element )
{
    int i, j;

    // skip initial space chars
    for ( i = 0; element[i] == ' '; ++i );

    for ( j = i; j < (int)(strlen(element)) && element[j] != ' '; ++j )
    {
        // make uppercase, offset to 0
        element[j - i] = toupper( element[j] );
    }
    // finalize the string
    element[j - i] = 0;
}


/************ from system_props.c *************/
real Get_Time( )
{
    struct timeval tim;

    gettimeofday( &tim, NULL );

    return ( tim.tv_sec + (tim.tv_usec / 1000000.0) );
}


real Get_Timing_Info( real t_start )
{
    struct timeval tim;
    real t_end;

    gettimeofday(&tim, NULL );
    t_end = tim.tv_sec + (tim.tv_usec / 1000000.0);

    return (t_end - t_start);
}


void Update_Timing_Info( real *t_start, real *timing )
{
    struct timeval tim;
    real t_end;

    gettimeofday(&tim, NULL );

    t_end = tim.tv_sec + (tim.tv_usec / 1000000.0);
    *timing += (t_end - *t_start);
    *t_start = t_end;
}


/*********** from io_tools.c **************/
int Get_Atom_Type( reax_interaction *reax_param, char *s )
{
    int i;

    for ( i = 0; i < reax_param->num_atom_types; ++i )
    {
        if ( !strcmp( reax_param->sbp[i].name, s ) )
        {
            return i;
        }
    }

    fprintf( stderr, "Unknown atom type %s. Terminating...\n", s );
    MPI_Abort( MPI_COMM_WORLD, UNKNOWN_ATOM_TYPE );

    return -1;
}


char *Get_Element( reax_system *system, int i )
{
    return &( system->reax_param.sbp[system->my_atoms[i].type].name[0] );
}


char *Get_Atom_Name( reax_system *system, int i )
{
    return &(system->my_atoms[i].name[0]);
}


void Allocate_Tokenizer_Space( char **line, char **backup, char ***tokens )
{
    int i;

    *line = (char*) smalloc( sizeof(char) * MAX_LINE, "Tokenizer:line" );

    *backup = (char*) smalloc( sizeof(char) * MAX_LINE, "Tokenizer:backup" );

    *tokens = (char**) smalloc( sizeof(char*) * MAX_TOKENS, "Tokenizer:tokens" );

    for ( i = 0; i < MAX_TOKENS; i++ )
    {
        (*tokens)[i] = (char*) smalloc(sizeof(char) * MAX_TOKEN_LEN, "Tokenizer:tokens[i]" );
    }
}


int Tokenize( const char* s, char*** tok )
{
    char test[MAX_LINE];
    char *sep = "\t \n!=";
    char *word;
    int count = 0;

    strncpy( test, s, MAX_LINE );

    for ( word = strtok(test, sep); word; word = strtok(NULL, sep) )
    {
        strncpy( (*tok)[count], word, MAX_LINE );
        count++;
    }

    return count;
}


/***************** taken from lammps ************************/
/* Safe wrapper around libc malloc
 *
 * n: num. of bytes to allocated
 * name: message with details about pointer, used for warnings/errors
 *
 * returns: ptr to allocated memory
 * */
void* smalloc( size_t n, const char *name )
{
    void *ptr;

    if ( n == 0 )
    {
        fprintf( stderr, "ERROR: failed to allocate %zu bytes for array %s.\n",
                n, name );
        MPI_Abort( MPI_COMM_WORLD, INSUFFICIENT_MEMORY );
    }

#if defined(DEBUG)
    fprintf( stderr, "INFO: requesting memory for %s\n", name );
#endif

    ptr = malloc( n );

    if ( ptr == NULL )
    {
        fprintf( stderr, "ERROR: failed to allocate %zu bytes for array %s.\n",
                n, name );
        MPI_Abort( MPI_COMM_WORLD, INSUFFICIENT_MEMORY );
    }

#if defined(DEBUG)
    fprintf( stderr, "INFO: successfuly assigned pointer for %s\n", name );
#endif

    return ptr;
}


/* Safe wrapper around libc realloc
 *
 * n: num. of bytes to reallocated
 * name: message with details about pointer, used for warnings/errors
 *
 * returns: ptr to reallocated memory
 * */
void* srealloc( void *ptr, size_t n, const char *name )
{
    void *new_ptr;

    if ( n == 0 )
    {
        fprintf( stderr, "ERROR: failed to reallocate %zu bytes for array %s.\n",
                n, name );
        MPI_Abort( MPI_COMM_WORLD, INSUFFICIENT_MEMORY );
    }

    if ( ptr == NULL )
    {
        fprintf( stderr, "INFO: trying to allocate %zu NEW bytes for array %s.\n",
                n, name );
    }

#if defined(DEBUG)
    fprintf( stderr, "INFO: requesting memory for %s\n", name );
#endif

    new_ptr = realloc( ptr, n );

    /* technically, ptr is still allocated and valid in this case,
     * but we needed more memory, so abort */
    if ( new_ptr == NULL )
    {
        fprintf( stderr, "ERROR: failed to reallocate %zu bytes for array %s.\n",
                n, name );
        MPI_Abort( MPI_COMM_WORLD, INSUFFICIENT_MEMORY );
    }

#if defined(DEBUG)
    fprintf( stderr, "INFO: successfuly assigned pointer for %s\n", name );
#endif

    return new_ptr;
}


/* Safe wrapper around libc calloc
 *
 * n: num. of elements to allocated (each of size bytes)
 * size: num. of bytes per element
 * name: message with details about pointer, used for warnings/errors
 *
 * returns: ptr to allocated memory, all bits initialized to zeros
 * */
void* scalloc( size_t n, size_t size, const char *name )
{
    void *ptr;

    if ( n == 0 )
    {
        fprintf( stderr, "ERROR: failed to allocate %zu bytes for array %s.\n",
                n * size, name );
        MPI_Abort( MPI_COMM_WORLD, INSUFFICIENT_MEMORY );
    }

#if defined(DEBUG)
    fprintf( stderr, "INFO: requesting memory for %s\n", name );
#endif

    ptr = calloc( n, size );

    if ( ptr == NULL )
    {
        fprintf( stderr, "ERROR: failed to allocate %zu bytes for array %s.\n",
                n * size, name );
        MPI_Abort( MPI_COMM_WORLD, INSUFFICIENT_MEMORY );
    }

#if defined(DEBUG)
    fprintf( stderr, "INFO: successfuly assigned pointer for %s\n", name );
#endif

    return ptr;
}


/* Safe wrapper around libc free
 *
 * ptr: pointer to dynamically allocated memory which will be deallocated
 * name: message with details about pointer, used for warnings/errors
 * */
void sfree( void *ptr, const char *name )
{
    if ( ptr == NULL )
    {
        fprintf( stderr, "WARNING: trying to free the already NULL pointer %s!\n",
                name );
        return;
    }

#if defined(DEBUG)
    fprintf( stderr, "INFO: freeing memory for %s\n", name );
#endif

    free( ptr );

#if defined(DEBUG)
    fprintf( stderr, "INFO: successfuly freed memory from pointer for %s\n", name );
#endif

    ptr = NULL;
}
